((nil . ((ellama-tools-edit-before-shell-commands
          . ((:command ".codex/hooks/pre-edit.sh")))
         (ellama-tools-edit-after-shell-commands
          . ((:command ".codex/hooks/post-edit.sh"
             :show-output t))))))
