;; Generated package description from ellama.el  -*- mode: lisp-data; no-byte-compile: t; lexical-binding:t -*-
(define-package "ellama" "1.32.1.0.20260812.1" "Tool for interacting with LLMs" '((emacs "28.1") (llm "0.31.1") (plz "0.8") (transient "0.7") (compat "29.1") (yaml "1.2.3")) :commit "a86e9154ae88907d3de2710195810cda986ef0d8" :authors '(("Sergey Kostyaev" . "sskostyaev@gmail.com")) :maintainer '("Sergey Kostyaev" . "sskostyaev@gmail.com") :keywords '("help" "local" "tools") :url "http://github.com/s-kostyaev/ellama")
